import { Either } from "./either.js";
import { normalizePathSegment, normalizeRest } from "./normalize.js";
export type Patterns = string | string[];
export type DefinitionResponse = [any, Patterns] | Patterns;
export type DefinitionValue = ((v: any) => DefinitionResponse) | DefinitionResponse;
export type Definition = Record<string, DefinitionValue>;
export type ValueFromDefinitionValue<V, PV> = V extends DefinitionValue ? V extends (v: any) => DefinitionResponse ? PV extends null ? Parameters<V>[0] : Parameters<V>[0] & PV : PV extends null ? Record<string, any> : PV : never;
export type Constructors<N extends string, D extends Definition, PV> = {
    [R in keyof D]: {
        (value: ValueFromDefinitionValue<D[R], PV>, config?: {
            rest?: string;
        }): {
            type: N;
            tag: R;
            context: RouteContext;
            value: ValueFromDefinitionValue<D[R], PV>;
        };
    } & {
        create<D2 extends Definition>(childDefinition: D2): Superoute<R extends string ? `${N}.${R}` : string, D2, ValueFromDefinitionValue<D[R], PV>>;
        ['superouter/metadata']: {
            value: ValueFromDefinitionValue<D[R], PV>;
            tag: R;
        };
    };
};
export type MatchOptions<D extends Definition, T, PV> = {
    [R in keyof D]: {
        (value: ValueFromDefinitionValue<D[R], PV>): T;
    };
};
export type RouteContext = {
    rest: string;
    patterns: string[];
    parentPatterns: string[];
};
export type API<N extends string, D extends Definition, PV> = {
    type: N;
    definition: {
        [R in keyof D]: D[R];
    };
    patterns: {
        [R in keyof D]: string[];
    };
    toPath(instance: InternalInstance<N, D, keyof D, PV>): string;
    toPathSafe(instance: InternalInstance<N, D, keyof D, PV>): Either<Error, string>;
    toLocalPath(instance: InternalInstance<N, D, keyof D, PV>): string;
    toPathSafe(instance: InternalInstance<N, D, keyof D, PV>): Either<Error, string>;
    fromPath(path: string): InternalInstance<N, D, keyof D, PV>;
    fromPathSafe(path: string): Either<Error, InternalInstance<N, D, keyof D, PV>>;
    fromLocalPath(path: string, defaults: PV): InternalInstance<N, D, keyof D, PV>;
    fromLocalPathSafe(path: string, defaults: PV): Either<Error, InternalInstance<N, D, keyof D, PV>>;
    fromLocalPathOr(fallback: () => InternalInstance<N, D, keyof D, PV>, path: string, defaults: PV): InternalInstance<N, D, keyof D, PV>;
    fromPathOr(fallback: () => InternalInstance<N, D, keyof D, PV>, path: string): InternalInstance<N, D, keyof D, PV>;
    match<T>(instance: InternalInstance<N, D, keyof D, PV>, options: MatchOptions<D, T, PV>): T;
    match<T>(options: MatchOptions<D, T, PV>): (instance: InternalInstance<N, D, keyof D, PV>) => T;
    def: <T>(fn: (x: InternalInstance<N, D, keyof D, PV>) => T) => (x: InternalInstance<N, D, keyof D, PV>) => T;
    parentValue: PV;
};
export type Is<R extends string> = `is${R}`;
export type Get<R extends string> = `get${R}`;
type Otherwise<D extends Definition> = {
    otherwise: <R extends Partial<keyof D>>(tags: R[]) => <T>(fn: () => T) => {
        [k in R]: () => T;
    };
};
type OtherwiseEmpty<D extends Definition> = {
    otherwise: () => <T>(fn: () => T) => {
        [k in keyof D]: () => T;
    };
};
type InternalInstance<N extends string, D extends Definition, K extends keyof D, PV> = ReturnType<Constructors<N, D, PV>[K]>;
export type Instance<A extends API<any, any, any>> = InternalInstance<A["type"], A["definition"], keyof A["definition"], A["parentValue"]>;
export type Superoute<N extends string, D extends Definition, PV> = Constructors<N, D, PV> & API<N, D, PV> & RouteIs<N, D, PV> & RouteGet<N, D, PV> & Otherwise<D> & OtherwiseEmpty<D>;
export type RouteIs<N extends string, D extends Definition, PV> = {
    [Key in keyof D as Is<Key extends string ? Key : never>]: (route: InternalInstance<N, D, keyof D, PV>) => boolean;
};
export type RouteGet<N extends string, D extends Definition, PV> = {
    [Key in keyof D as Get<Key extends string ? Key : never>]: <T>(fallback: T, visitor: (value: ValueFromDefinitionValue<D[Key], PV>) => T, route: InternalInstance<N, D, keyof D, PV>) => T;
};
export type AnyRoute = {
    type: string;
    tag: string;
    value: Record<string, any>;
    context: RouteContext;
};
declare function toPathSafeExternal(route: AnyRoute): Either<Error, string>;
export { toPathSafeExternal as toPathSafe };
declare function toPathOrExternal(otherwise: string, route: AnyRoute): string;
export { toPathOrExternal as toPathOr };
declare function toPathExternal(route: AnyRoute): string;
export { toPathExternal as toPath };
export declare function create<D extends Definition, N extends string = "Main">(definition: D): Superoute<N, D, null>;
export declare function create<D extends Definition, N extends string>(name: N, definition: D): Superoute<N, D, null>;
export type Tag<A> = A extends API<any, any, any> ? Instance<A>["tag"] : A extends Instance<any> ? A["tag"] : A extends {
    ['superouter/metadata']: any;
} ? A["superouter/metadata"]["value"] : never;
export type Value<A> = A extends API<any, any, any> ? Instance<A>["value"] : A extends Instance<any> ? A["value"] : A extends {
    ['superouter/metadata']: any;
} ? A["superouter/metadata"]["value"] : never;
export { normalizePathSegment, normalizeRest };
