// lib/either.ts
var Either = {
  Left(value) {
    return { type: "Either", tag: "Left", value };
  },
  Right(value) {
    return { type: "Either", tag: "Right", value };
  }
};

// lib/normalize.ts
function normalizePathSegment(s) {
  if (s == "" || s == "/") {
    return "/";
  } else if (s.startsWith("/") && s.endsWith("/")) {
    return s.slice(0, -1);
  } else if (!s.startsWith("/") && s.endsWith("/")) {
    return "/" + s.slice(0, -1);
  } else {
    return s;
  }
}
function normalizeRest(s) {
  if (!s) {
    s = "";
  }
  if (s.at(0) == "/") {
    s = s.slice(1);
  }
  if (s.at(-1) == "/") {
    s = s.slice(0, -1);
  }
  return s;
}

// lib/fromPath.ts
function safe(_path, _pattern) {
  if (_path == null) {
    throw new Error("Provided path was null but must be a URL path");
  }
  if (_pattern == null) {
    throw new Error("Provided pattern was null but must be a URL pattern");
  }
  const path = _path + "/";
  const pattern = _pattern + "/";
  let pathI = 0;
  let patternI = 0;
  const value = {};
  let varName = "";
  let varValue = "";
  let literalName = "";
  let literalValue = "";
  let rest = "/";
  let mode = "initialize";
  let prevMode = "initialize";
  let score = 0;
  const maxIterations = path.length + pattern.length;
  let iterations = 0;
  while (true) {
    if (mode == "initialize") {
      while (pattern[patternI] === "/") {
        patternI++;
        continue;
      }
      while (path[pathI] === "/") {
        pathI++;
        continue;
      }
      if (prevMode == "collectLiteralName") {
        mode = "collectLiteralValue";
        literalValue = "";
        continue;
      }
      if (prevMode == "collectVarName") {
        if (varName in value) {
          return {
            type: "Either",
            tag: "Left",
            value: new Error(`Pattern has duplicated varName ':${varName}'`)
          };
        }
        score += 2;
        mode = "collectVarValue";
        varValue = "";
        continue;
      }
      if (prevMode == "collectLiteralValue") {
        if (literalName !== literalValue) {
          return {
            type: "Either",
            tag: "Left",
            value: new Error(
              `Expected literal path segment '/${literalName}' but instead found '/${literalValue}'`
            )
          };
        }
        score += 4;
        literalName = "";
        literalValue = "";
      }
      if (prevMode == "collectVarValue") {
        if (!varValue) {
          return {
            type: "Either",
            tag: "Left",
            value: new Error(
              `Expected binding for path variable ':${varName}' but instead found nothing'`
            )
          };
        }
        value[varName] = varValue;
        varName = "";
        varValue = "";
      }
      if (pattern[patternI] === ":") {
        mode = "collectVarName";
        varName = "";
        patternI++;
      } else if (pattern[patternI] != null) {
        mode = "collectLiteralName";
        literalName = "";
      } else if (path[pathI] != null) {
        mode = "collectRest";
      } else {
        break;
      }
    }
    if (iterations >= maxIterations) {
      throw new Error(
        `Unexpected recursive logic while parsing path '${_path}' using pattern '${_pattern}'`
      );
    }
    iterations++;
    if ((mode == "collectLiteralName" || mode == "collectVarName") && pattern[patternI] === "/") {
      prevMode = mode;
      mode = "initialize";
      continue;
    }
    if ((mode == "collectLiteralValue" || mode == "collectVarValue" || mode == "collectRest") && path[pathI] === "/") {
      prevMode = mode;
      mode = "initialize";
      continue;
    }
    if ((mode == "collectLiteralName" || mode == "collectVarName") && pattern[patternI] == null) {
      prevMode = mode;
      mode = "initialize";
      continue;
    }
    if ((mode == "collectLiteralValue" || mode == "collectVarValue" || mode == "collectRest") && path[pathI] == null) {
      prevMode = mode;
      mode = "initialize";
      continue;
    }
    if (mode == "collectLiteralName") {
      literalName += pattern[patternI];
      patternI++;
      continue;
    }
    if (mode == "collectLiteralValue") {
      literalValue += path[pathI];
      pathI++;
      continue;
    }
    if (mode == "collectVarName") {
      varName += pattern[patternI];
      patternI++;
      continue;
    }
    if (mode == "collectVarValue") {
      varValue += path[pathI];
      pathI++;
      continue;
    }
    if (mode == "collectRest") {
      while (path[pathI + 1] != null) {
        rest += path[pathI];
        pathI++;
      }
      break;
    }
  }
  if (rest) {
    score = Math.max(0, score - 1);
  }
  return { type: "Either", tag: "Right", value: { rest: normalizeRest(rest), value, score } };
}

// lib/index.ts
function match(...args) {
  if (args.length == 2) {
    const [instance, options] = args;
    return options[instance.tag](instance.value);
  } else if (args.length == 1) {
    const [options] = args;
    return (instance) => {
      return options[instance.tag](instance.value);
    };
  } else {
    throw new Error("Expected either 2 or 1 parameters but received: " + args.length);
  }
}
function def(fn) {
  return (x) => fn(x);
}
function otherwise(tags) {
  return (fn) => Object.fromEntries(tags.map((tag) => [tag, fn]));
}
function toPathSafeExternal(route) {
  return toPathInternalSafe(route, route.context.parentPatterns, route.context.patterns);
}
function toPathOrExternal(otherwise2, route) {
  const answer = toPathInternalSafe(route, route.context.parentPatterns, route.context.patterns);
  if (answer.tag === "Left") {
    return otherwise2;
  } else {
    return answer.value;
  }
}
function toPathExternal(route) {
  const answer = toPathInternalSafe(route, route.context.parentPatterns, route.context.patterns);
  if (answer.tag === "Left") {
    throw answer.value;
  } else {
    return answer.value;
  }
}
function toPathInternalSafe(route, parentPatterns, localPatterns) {
  const errors = [];
  const paths = [];
  const ranks = {};
  let bestPath = null;
  let bestRank = 0;
  let error = null;
  const pathPrefix = [];
  for (const parentPattern of parentPatterns) {
    for (const segment of parentPattern.split("/")) {
      if (segment.startsWith(":")) {
        const name = segment.slice(1);
        if (route.value[name]) {
          pathPrefix.push(route.value[name]);
        } else {
          error = `Could not build pattern ${parentPattern} from value ${JSON.stringify(
            route.value
          )} as '${name} was empty'`;
          errors.push(error);
          break;
        }
      } else {
        pathPrefix.push(segment);
      }
    }
  }
  for (const _pattern of localPatterns) {
    const pattern = normalizePathSegment(_pattern);
    let path = [];
    let rank = 0;
    for (const segment of pattern.split("/")) {
      if (segment.startsWith(":")) {
        const name = segment.slice(1);
        if (route.value[name]) {
          path.push(route.value[name]);
          rank += 2;
        } else {
          error = `Could not build pattern ${pattern} from value ${JSON.stringify(
            route.value
          )} as '${name} was empty'`;
          errors.push(error);
          path = null;
          break;
        }
      } else {
        path.push(segment);
        rank += 4;
      }
    }
    if (path != null) {
      paths.push(path);
      ranks[pattern] = rank;
      if (rank > bestRank || bestPath == null) {
        bestRank = rank;
        bestPath = path;
      }
      continue;
    }
  }
  if (bestPath) {
    const value = (pathPrefix.length ? normalizePathSegment(pathPrefix.join("/")) : "") + normalizePathSegment(bestPath.concat(route.context.rest).join("/"));
    return { type: "Either", tag: "Right", value };
  } else if (errors.length) {
    return { type: "Either", tag: "Left", value: new Error(errors[0]) };
  } else {
    throw new Error(`Unexpected failure converting route ${route} to url`);
  }
}
function toPath(route) {
  const result = toPathSafe(route);
  if (result.tag === "Left") {
    throw result.value;
  } else {
    return result.value;
  }
}
function toLocalPath(route) {
  const result = toPathInternalSafe(route, [], route.context.patterns);
  if (result.tag === "Left") {
    throw result.value;
  } else {
    return result.value;
  }
}
function toPathSafe(route) {
  const answer = toPathInternalSafe(route, route.context.parentPatterns, route.context.patterns);
  if (answer.tag === "Right") {
    return Either.Right(normalizePathSegment(answer.value));
  }
  return answer;
}
function internalCreate(type, routes, routeContext, parentPatterns) {
  function fromLocalPathOr(otherwise2, path, defaults) {
    const res = fromLocalPathSafe(path, defaults);
    if (res.tag == "Left") {
      return otherwise2();
    } else {
      return res.value;
    }
  }
  function fromPathOr(otherwise2, path) {
    const res = fromLocalPathSafe(path, {});
    if (res.tag == "Left") {
      return otherwise2();
    } else {
      return res.value;
    }
  }
  function fromPathSafe(path) {
    return fromPathSafeInternal(path, parentPatterns, api.patterns, {});
  }
  function fromPath(path) {
    const res = fromPathSafe(path);
    if (res.tag === "Left") {
      throw res.value;
    } else {
      return res.value;
    }
  }
  function fromLocalPathSafe(path, defaults) {
    return fromPathSafeInternal(path, [], api.patterns, defaults);
  }
  function fromPathSafeInternal(path, parentPatterns2, localPatterns, defaults) {
    if (path.includes("://")) {
      return Either.Left(
        new Error(
          `Please provide a path segment instead of a complete url found:'${path}'`
        )
      );
    }
    let bestRank = 0;
    let bestParentMatch = {
      value: {},
      rest: path,
      score: -1
    };
    let lastParsedParentPath;
    for (const parentPattern of parentPatterns2) {
      const parsedPrefix = safe(path, parentPattern);
      lastParsedParentPath = parsedPrefix;
      if (parsedPrefix.tag === "Left") {
        continue;
      }
      if (parsedPrefix.value.score > bestRank) {
        bestParentMatch = parsedPrefix.value;
        bestRank = parsedPrefix.value.score;
      }
    }
    if (parentPatterns2.length && bestParentMatch.score === -1) {
      return lastParsedParentPath;
    }
    const { value: parentValue = {}, rest } = bestParentMatch;
    path = rest;
    bestRank = 0;
    let bestRoute = null;
    let error = null;
    for (const [tag, patterns] of Object.entries(localPatterns)) {
      for (const pattern of patterns) {
        const result = safe(path, normalizePathSegment(pattern));
        if (result.tag === "Left") {
          if (error == null) {
            error = result;
          }
        } else {
          if (bestRoute == null || result.value.score > bestRank) {
            bestRoute = api[tag](
              {
                ...result.value.value
              },
              { rest: result.value.rest }
            );
            bestRank = result.value.score;
          }
        }
      }
    }
    if (bestRoute == null) {
      if (error) {
        return error;
      } else {
        return {
          type: "Either",
          tag: "Left",
          value: new Error(
            `Could not parse url ${path} into any pattern on type ${type}`
          )
        };
      }
    } else {
      bestRoute.value = Object.assign(
        {},
        defaults,
        bestRoute.value,
        parentPatterns2.length ? parentValue : {}
      );
      return {
        type: "Either",
        tag: "Right",
        value: bestRoute
      };
    }
  }
  function fromLocalPath(route, defaults) {
    const res = fromLocalPathSafe(route, defaults);
    if (res.tag == "Left") {
      throw res.value;
    } else {
      return res.value;
    }
  }
  const api = {
    type,
    patterns: {},
    definition: routes,
    toPath,
    toPathSafe,
    fromPath,
    fromPathSafe,
    toLocalPath,
    toLocalPathSafe: toPathInternalSafe,
    fromLocalPath,
    fromLocalPathSafe,
    fromPathOr,
    fromLocalPathOr,
    match,
    def,
    otherwise: (...args) => {
      if (args.length === 0) {
        return otherwise(Object.keys(routes));
      }
      return otherwise(args[0]);
    }
  };
  for (const [tag, dv] of Object.entries(routes)) {
    const of = typeof dv == "function" ? dv : () => dv;
    api[tag] = (_value = {}, config) => {
      const { ...value } = _value;
      const context = {
        // prefix: normalizePathSegment(routeContext.prefix),
        rest: normalizeRest(config?.rest ?? ""),
        parentPatterns,
        patterns: api.patterns[tag]
      };
      return { type, tag, value: { ...value }, context };
    };
    api[tag].create = (definition) => {
      return internalCreate(
        `${type}.${tag}`,
        definition,
        routeContext,
        (parentPatterns.length ? parentPatterns : [""]).flatMap(
          (x) => api.patterns[tag].map((childPattern) => x + childPattern)
        )
      );
    };
    const res = of({});
    let patterns = [];
    if (typeof res === "string") {
      patterns = [res];
    } else if (Array.isArray(res)) {
      if (Array.isArray(res[1])) {
        patterns = res[1];
      } else {
        patterns = res;
      }
    }
    if (patterns.length == 0) {
      throw new Error(`Must provide at least path pattern for ${type}.${tag}`);
    }
    api.patterns[tag] = patterns;
    api[`is${tag}`] = (v) => v.tag === tag;
    api[`get${tag}`] = (fallback, f, v) => v.tag === tag ? f(v.value) : fallback;
  }
  return api;
}
function create(...args) {
  if (typeof args[0] === "string") {
    return internalCreate(args[0], args[1], { rest: "", patterns: [], parentPatterns: [] }, []);
  } else {
    return internalCreate("Main", args[0], { rest: "", patterns: [], parentPatterns: [] }, []);
  }
}
export {
  create,
  normalizePathSegment,
  normalizeRest,
  toPathExternal as toPath,
  toPathOrExternal as toPathOr,
  toPathSafeExternal as toPathSafe
};
//# sourceMappingURL=superouter.esm.js.map
